import asyncio
import importlib
from typing import List, Type, Union, TYPE_CHECKING

from motor.motor_asyncio import AsyncIOMotorDatabase

if TYPE_CHECKING:
    from beanie.odm.documents import DocType


def get_model(dot_path: str) -> Type["DocType"]:
    """
    Get the model by the path in format bar.foo.Model

    :param dot_path: str - dot seprated path to the model
    :return: Type[DocType] - class of the model
    """
    try:
        module_name, class_name = dot_path.rsplit(".", 1)
        return getattr(importlib.import_module(module_name), class_name)

    except ValueError:
        raise ValueError(
            f"'{dot_path}' doesn't have '.' path, eg. path.to.your.model.class"
        )

    except AttributeError:
        raise AttributeError(
            f"module '{module_name}' has no class called '{class_name}'"
        )


async def init_beanie(
    database: AsyncIOMotorDatabase,
    document_models: List[Union[Type["DocType"], str]],
    allow_index_dropping: bool = True,
):
    """
    Beanie initialization

    :param database: AsyncIOMotorDatabase - motor database instance
    :param document_models: List[Union[Type[DocType], str]] - model classes
    or strings with dot separated paths
    :param allow_index_dropping: bool - if index dropping is allowed.
    Default True
    :return: None
    """
    collection_inits = []
    for model in document_models:
        if isinstance(model, str):
            model = get_model(model)
        collection_inits.append(
            model.init_collection(
                database, allow_index_dropping=allow_index_dropping
            )
        )

    await asyncio.gather(*collection_inits)



from typing import (
    List,
    Type,
    TYPE_CHECKING,
    Optional,
    Union,
    Mapping,
    Any,
    Dict,
)

from pymongo.client_session import ClientSession
from pymongo.results import UpdateResult

from beanie.odm.interfaces.session import SessionMethods
from beanie.odm.interfaces.update import (
    UpdateMethods,
)
from beanie.odm.operators.update import BaseUpdateOperator

if TYPE_CHECKING:
    from beanie.odm.documents import DocType


class UpdateQuery(UpdateMethods, SessionMethods):
    """
    Update Query base class

    Inherited from:

    - [SessionMethods](https://roman-right.github.io/beanie/api/interfaces/#sessionmethods)
    - [UpdateMethods](https://roman-right.github.io/beanie/api/interfaces/#aggregatemethods)
    """

    def __init__(
        self,
        document_model: Type["DocType"],
        find_query: Union[Dict[str, Any], Mapping[str, Any]],
    ):
        self.document_model = document_model
        self.find_query = find_query
        self.update_expressions: List[
            Union[Dict[str, Any], Mapping[str, Any]]
        ] = []
        self.session = None

    @property
    def update_query(self) -> Dict[str, Any]:
        query: Dict[str, Any] = {}
        for expression in self.update_expressions:
            if isinstance(expression, BaseUpdateOperator):
                query.update(expression.query)
            elif isinstance(expression, dict):
                query.update(expression)
            else:
                raise TypeError("Wrong expression type")
        return query

    def update(
        self,
        *args: Union[Dict[str, Any], Mapping[str, Any]],
        session: Optional[ClientSession] = None
    ) -> "UpdateQuery":
        """
        Provide modifications to the update query. The same as `update()`

        :param args: *Union[dict, Mapping] - the modifications to apply.
        :param session: Optional[ClientSession]
        :return: UpdateMany query
        """
        self.set_session(session=session)
        self.update_expressions += args
        return self


class UpdateMany(UpdateQuery):
    """
    Update Many query class

    Inherited from:

    - [UpdateQuery](https://roman-right.github.io/beanie/api/queries/#updatequery)
    """

    def update_many(
        self,
        *args: Union[Dict[str, Any], Mapping[str, Any]],
        session: Optional[ClientSession] = None
    ):
        """
        Provide modifications to the update query

        :param args: *Union[dict, Mapping] - the modifications to apply.
        :param session: Optional[ClientSession]
        :return: UpdateMany query
        """
        return self.update(*args, session=session)

    def __await__(self) -> UpdateResult:
        """
        Run the query
        :return:
        """
        yield from self.document_model.get_motor_collection().update_many(
            self.find_query, self.update_query, session=self.session
        )


class UpdateOne(UpdateQuery):
    """
    Update One query class

    Inherited from:

    - [UpdateQuery](https://roman-right.github.io/beanie/api/queries/#updatequery)
    """

    def update_one(
        self,
        *args: Union[Dict[str, Any], Mapping[str, Any]],
        session: Optional[ClientSession] = None
    ):
        """
        Provide modifications to the update query. The same as `update()`

        :param args: *Union[dict, Mapping] - the modifications to apply.
        :param session: Optional[ClientSession]
        :return: UpdateMany query
        """
        return self.update(*args, session=session)

    def __await__(self) -> UpdateResult:
        """
        Run the query
        :return:
        """
        yield from self.document_model.get_motor_collection().update_one(
            self.find_query, self.update_query, session=self.session
        )



from abc import abstractmethod
from typing import Type, Any, Optional, Union, List, Dict

from pydantic import BaseModel
from pymongo.client_session import ClientSession

from beanie.odm.fields import ExpressionField
from beanie.odm.queries.aggregation import AggregationQuery


class AggregateMethods:
    """
    Aggregate methods
    """

    @abstractmethod
    def aggregate(
        self,
        aggregation_pipeline,
        projection_model: Type[BaseModel] = None,
        session: Optional[ClientSession] = None,
    ) -> AggregationQuery:
        ...

    async def sum(
        self,
        field: Union[str, ExpressionField],
        session: Optional[ClientSession] = None,
    ) -> float:
        """
        Sum of values of the given field

        Example:

        ```python

        class Sample(Document):
            price: int
            count: int

        sum_count = await Document.find(Sample.price <= 100).sum(Sample.count)

        ```

        :param field: Union[str, ExpressionField]
        :param session: Optional[ClientSession] - pymongo session
        :return: float - sum
        """
        pipeline = [
            {"$group": {"_id": None, "sum": {"$sum": f"${field}"}}},
            {"$project": {"_id": 0, "sum": 1}},
        ]

        result: List[Dict[str, Any]] = await self.aggregate(
            aggregation_pipeline=pipeline, session=session
        ).to_list()
        return result[0]["sum"]

    async def avg(
        self, field, session: Optional[ClientSession] = None
    ) -> float:
        """
        Average of values of the given field

        Example:

        ```python

        class Sample(Document):
            price: int
            count: int

        avg_count = await Document.find(Sample.price <= 100).avg(Sample.count)
        ```

        :param field: Union[str, ExpressionField]
        :param session: Optional[ClientSession] - pymongo session
        :return: float - avg
        """
        pipeline = [
            {"$group": {"_id": None, "avg": {"$avg": f"${field}"}}},
            {"$project": {"_id": 0, "avg": 1}},
        ]

        result: List[Dict[str, Any]] = await self.aggregate(
            aggregation_pipeline=pipeline, session=session
        ).to_list()
        return result[0]["avg"]

    async def max(
        self,
        field: Union[str, ExpressionField],
        session: Optional[ClientSession] = None,
    ) -> Any:
        """
        Max of the values of the given field

        Example:

        ```python

        class Sample(Document):
            price: int
            count: int

        max_count = await Document.find(Sample.price <= 100).max(Sample.count)
        ```

        :param field: Union[str, ExpressionField]
        :param session: Optional[ClientSession] - pymongo session
        :return: float - max
        """
        pipeline = [
            {"$group": {"_id": None, "max": {"$max": f"${field}"}}},
            {"$project": {"_id": 0, "max": 1}},
        ]

        result: List[Dict[str, Any]] = await self.aggregate(
            aggregation_pipeline=pipeline, session=session
        ).to_list()
        return result[0]["max"]

    async def min(
        self,
        field: Union[str, ExpressionField],
        session: Optional[ClientSession] = None,
    ) -> Any:
        """
        Min of the values of the given field

        Example:

        ```python

        class Sample(Document):
            price: int
            count: int

        min_count = await Document.find(Sample.price <= 100).min(Sample.count)
        ```

        :param field: Union[str, ExpressionField]
        :param session: Optional[ClientSession] - pymongo session
        :return: float - max
        """
        pipeline = [
            {"$group": {"_id": None, "min": {"$min": f"${field}"}}},
            {"$project": {"_id": 0, "min": 1}},
        ]

        result: List[Dict[str, Any]] = await self.aggregate(
            aggregation_pipeline=pipeline, session=session
        ).to_list()
        return result[0]["min"]



from typing import (
    Optional,
    List,
    Type,
    Union,
    Tuple,
    Mapping,
    TypeVar,
    Dict,
    Any,
)

from bson import ObjectId
from motor.motor_asyncio import AsyncIOMotorDatabase, AsyncIOMotorCollection
from pydantic import ValidationError
from pydantic.main import BaseModel
from pymongo.client_session import ClientSession
from pymongo.results import (
    DeleteResult,
    InsertOneResult,
    InsertManyResult,
)

from beanie.exceptions import (
    CollectionWasNotInitialized,
    ReplaceError,
    DocumentNotFound,
)
from beanie.odm.enums import SortDirection
from beanie.odm.fields import PydanticObjectId, ExpressionField
from beanie.odm.interfaces.update import (
    UpdateMethods,
)
from beanie.odm.models import (
    InspectionResult,
    InspectionStatuses,
    InspectionError,
)
from beanie.odm.operators.find.comparison import In
from beanie.odm.queries.aggregation import AggregationQuery
from beanie.odm.queries.find import FindOne, FindMany
from beanie.odm.queries.update import UpdateMany
from beanie.odm.utils.collection import collection_factory
from beanie.odm.utils.dump import get_dict

DocType = TypeVar("DocType", bound="Document")


class Document(BaseModel, UpdateMethods):
    """
    Document Mapping class.

    Fields:

    - `id` - MongoDB document ObjectID "_id" field.
    Mapped to the PydanticObjectId class

    Inherited from:

    - Pydantic BaseModel
    - [UpdateMethods](https://roman-right.github.io/beanie/api/interfaces/#aggregatemethods)
    """

    id: Optional[PydanticObjectId] = None

    def __init__(self, *args, **kwargs):
        super(Document, self).__init__(*args, **kwargs)
        self.get_motor_collection()

    async def _sync(self) -> None:
        """
        Update local document from the database
        :return: None
        """
        if self.id is None:
            raise ValueError("Document has no id")
        new_instance: Document = await self.get(self.id)
        for key, value in dict(new_instance).items():
            setattr(self, key, value)

    async def insert(
        self: DocType, session: Optional[ClientSession] = None
    ) -> DocType:
        """
        Insert the document (self) to the collection
        :return: Document
        """
        result = await self.get_motor_collection().insert_one(
            get_dict(self), session=session
        )
        new_id = result.inserted_id
        if not isinstance(new_id, self.__fields__["id"].type_):
            new_id = self.__fields__["id"].type_(new_id)
        self.id = new_id
        return self

    async def create(
        self: DocType, session: Optional[ClientSession] = None
    ) -> DocType:
        """
        The same as self.insert()
        :return: Document
        """
        return await self.insert(session=session)

    @classmethod
    async def insert_one(
        cls: Type[DocType],
        document: DocType,
        session: Optional[ClientSession] = None,
    ) -> InsertOneResult:
        """
        Insert one document to the collection
        :param document: Document - document to insert
        :param session: ClientSession - pymongo session
        :return: InsertOneResult
        """
        return await cls.get_motor_collection().insert_one(
            get_dict(document), session=session
        )

    @classmethod
    async def insert_many(
        cls: Type[DocType],
        documents: List[DocType],
        session: Optional[ClientSession] = None,
    ) -> InsertManyResult:

        """
        Insert many documents to the collection

        :param documents:  List["Document"] - documents to insert
        :param session: ClientSession - pymongo session
        :return: InsertManyResult
        """
        documents_list = [get_dict(document) for document in documents]
        return await cls.get_motor_collection().insert_many(
            documents_list,
            session=session,
        )

    @classmethod
    async def get(
        cls: Type[DocType],
        document_id: PydanticObjectId,
        session: Optional[ClientSession] = None,
    ) -> Optional[DocType]:
        """
        Get document by id

        :param document_id: PydanticObjectId - document id
        :param session: Optional[ClientSession] - pymongo session
        :return: Union["Document", None]
        """
        return await cls.find_one({"_id": document_id}, session=session)

    @classmethod
    def find_one(
        cls,
        *args: Union[Dict[str, Any], Mapping[str, Any], bool],
        projection_model: Optional[Type[BaseModel]] = None,
        session: Optional[ClientSession] = None,
    ) -> FindOne:
        """
        Find one document by criteria.
        Returns [FindOne](https://roman-right.github.io/beanie/api/queries/#findone) query object

        :param args: *Union[Dict[str, Any],
        Mapping[str, Any], bool] - search criteria
        :param projection_model: Optional[Type[BaseModel]] - projection model
        :param session: Optional[ClientSession] - pymongo session instance
        :return: [FindOne](https://roman-right.github.io/beanie/api/queries/#findone) - find query instance
        """
        return FindOne(document_model=cls).find_one(
            *args,
            projection_model=projection_model,
            session=session,
        )

    @classmethod
    def find_many(
        cls,
        *args: Union[Dict[str, Any], Mapping[str, Any], bool],
        skip: Optional[int] = None,
        limit: Optional[int] = None,
        sort: Union[None, str, List[Tuple[str, SortDirection]]] = None,
        projection_model: Optional[Type[BaseModel]] = None,
        session: Optional[ClientSession] = None,
    ) -> FindMany:
        """
        Find many documents by criteria.
        Returns [FindMany](https://roman-right.github.io/beanie/api/queries/#findmany) query object

        :param args: *Union[Dict[str, Any],
        Mapping[str, Any], bool] - search criteria
        :param skip: Optional[int] - The number of documents to omit.
        :param limit: Optional[int] - The maximum number of results to return.
        :param sort: Union[None, str, List[Tuple[str, SortDirection]]] - A key
        or a list of (key, direction) pairs specifying the sort order
        for this query.
        :param projection_model: Optional[Type[BaseModel]] - projection model
        :param session: Optional[ClientSession] - pymongo session
        :return: [FindMany](https://roman-right.github.io/beanie/api/queries/#findmany) - query instance
        """
        return FindMany(document_model=cls).find_many(
            *args,
            sort=sort,
            skip=skip,
            limit=limit,
            projection_model=projection_model,
            session=session,
        )

    @classmethod
    def find(
        cls,
        *args: Union[Dict[str, Any], Mapping[str, Any], bool],
        skip: Optional[int] = None,
        limit: Optional[int] = None,
        sort: Union[None, str, List[Tuple[str, SortDirection]]] = None,
        projection_model: Optional[Type[BaseModel]] = None,
        session: Optional[ClientSession] = None,
    ) -> FindMany:
        """
        The same as find_many
        """
        return cls.find_many(
            *args,
            skip=skip,
            limit=limit,
            sort=sort,
            projection_model=projection_model,
            session=session,
        )

    @classmethod
    def find_all(
        cls,
        skip: Optional[int] = None,
        limit: Optional[int] = None,
        sort: Union[None, str, List[Tuple[str, SortDirection]]] = None,
        projection_model: Optional[Type[BaseModel]] = None,
        session: Optional[ClientSession] = None,
    ) -> FindMany:
        """
        Get all the documents

        :param skip: Optional[int] - The number of documents to omit.
        :param limit: Optional[int] - The maximum number of results to return.
        :param sort: Union[None, str, List[Tuple[str, SortDirection]]] - A key
        or a list of (key, direction) pairs specifying the sort order
        for this query.
        :param projection_model: Optional[Type[BaseModel]] - projection model
        :param session: Optional[ClientSession] - pymongo session
        :return: [FindMany](https://roman-right.github.io/beanie/api/queries/#findmany) - query instance
        """
        return cls.find_many(
            {},
            skip=skip,
            limit=limit,
            sort=sort,
            projection_model=projection_model,
            session=session,
        )

    @classmethod
    def all(
        cls,
        skip: Optional[int] = None,
        limit: Optional[int] = None,
        sort: Union[None, str, List[Tuple[str, SortDirection]]] = None,
        projection_model: Optional[Type[BaseModel]] = None,
        session: Optional[ClientSession] = None,
    ) -> FindMany:
        """
        the same as find_all
        """
        return cls.find_all(
            skip=skip,
            limit=limit,
            sort=sort,
            projection_model=projection_model,
            session=session,
        )

    async def replace(
        self, session: Optional[ClientSession] = None
    ) -> DocType:
        """
        Fully update the document in the database

        :param session: Optional[ClientSession] - pymongo session.
        :return: None
        """
        if self.id is None:
            raise ValueError("Document must have an id")

        await self.find_one({"_id": self.id}).replace_one(
            self, session=session
        )
        return self

    async def save(self, session: Optional[ClientSession] = None) -> DocType:
        """
        Update an existing model in the database or insert it if it does not yet exist.

        :param session: Optional[ClientSession] - pymongo session.
        :return: None
        """

        try:
            return await self.replace(session=session)
        except (ValueError, DocumentNotFound):
            return await self.insert(session=session)

    @classmethod
    async def replace_many(
        cls: Type[DocType],
        documents: List[DocType],
        session: Optional[ClientSession] = None,
    ) -> None:
        """
        Replace list of documents

        :param documents: List["Document"]
        :param session: Optional[ClientSession] - pymongo session.
        :return: None
        """
        ids_list = [document.id for document in documents]
        if await cls.find(In(cls.id, ids_list)).count() != len(ids_list):
            raise ReplaceError(
                "Some of the documents are not exist in the collection"
            )
        await cls.find(In(cls.id, ids_list), session=session).delete()
        await cls.insert_many(documents, session=session)

    async def update(
        self, *args, session: Optional[ClientSession] = None
    ) -> None:
        """
        Partially update the document in the database

        :param args: *Union[dict, Mapping] - the modifications to apply.
        :param session: ClientSession - pymongo session.
        :return: None
        """
        await self.find_one({"_id": self.id}).update(*args, session=session)
        await self._sync()

    @classmethod
    def update_all(
        cls,
        *args: Union[dict, Mapping],
        session: Optional[ClientSession] = None,
    ) -> UpdateMany:
        """
        Partially update all the documents

        :param args: *Union[dict, Mapping] - the modifications to apply.
        :param session: ClientSession - pymongo session.
        :return: UpdateMany query
        """
        return cls.find_all().update_many(*args, session=session)

    async def delete(
        self, session: Optional[ClientSession] = None
    ) -> DeleteResult:
        """
        Delete the document

        :param session: Optional[ClientSession] - pymongo session.
        :return: DeleteResult - pymongo DeleteResult instance.
        """
        return await self.find_one({"_id": self.id}).delete(session=session)

    @classmethod
    async def delete_all(
        cls, session: Optional[ClientSession] = None
    ) -> DeleteResult:
        """
        Delete all the documents

        :param session: Optional[ClientSession] - pymongo session.
        :return: DeleteResult - pymongo DeleteResult instance.
        """
        return await cls.find_all().delete(session=session)

    @classmethod
    def aggregate(
        cls,
        aggregation_pipeline: list,
        projection_model: Type[BaseModel] = None,
        session: Optional[ClientSession] = None,
    ) -> AggregationQuery:
        """
        Aggregate over collection.
        Returns [AggregationQuery](https://roman-right.github.io/beanie/api/queries/#aggregationquery) query object
        :param aggregation_pipeline: list - aggregation pipeline
        :param projection_model: Type[BaseModel]
        :param session: Optional[ClientSession]
        :return: [AggregationQuery](https://roman-right.github.io/beanie/api/queries/#aggregationquery)
        """
        return cls.find_all().aggregate(
            aggregation_pipeline=aggregation_pipeline,
            projection_model=projection_model,
            session=session,
        )

    @classmethod
    async def count(cls) -> int:
        """
        Number of documents in the collections
        The same as find_all().count()

        :return: int
        """
        return await cls.find_all().count()

    @classmethod
    async def init_collection(
        cls, database: AsyncIOMotorDatabase, allow_index_dropping: bool
    ) -> None:
        """
        Internal CollectionMeta class creator

        :param database: AsyncIOMotorDatabase - motor database instance
        :param allow_index_dropping: bool - if index dropping is allowed
        :return: None
        """
        collection_class = getattr(cls, "Collection", None)
        collection_meta = await collection_factory(
            database=database,
            document_model=cls,
            allow_index_dropping=allow_index_dropping,
            collection_class=collection_class,
        )
        setattr(cls, "CollectionMeta", collection_meta)

        for k, v in cls.__fields__.items():
            path = v.alias or v.name
            setattr(cls, k, ExpressionField(path))

    @classmethod
    def _get_collection_meta(cls) -> Type:
        """
        Get internal CollectionMeta class, which was created on
        the collection initialization step

        :return: CollectionMeta class
        """
        collection_meta = getattr(cls, "CollectionMeta", None)
        if collection_meta is None:
            raise CollectionWasNotInitialized
        return collection_meta

    @classmethod
    def get_motor_collection(cls) -> AsyncIOMotorCollection:
        """
        Get Motor Collection to access low level control

        :return: AsyncIOMotorCollection
        """
        collection_meta = cls._get_collection_meta()
        return collection_meta.motor_collection

    @classmethod
    async def inspect_collection(
        cls, session: Optional[ClientSession] = None
    ) -> InspectionResult:
        """
        Check, if documents, stored in the MongoDB collection
        are compatible with the Document schema

        :return: InspectionResult
        """
        inspection_result = InspectionResult()
        async for json_document in cls.get_motor_collection().find(
            {}, session=session
        ):
            try:
                cls.parse_obj(json_document)
            except ValidationError as e:
                if inspection_result.status == InspectionStatuses.OK:
                    inspection_result.status = InspectionStatuses.FAIL
                inspection_result.errors.append(
                    InspectionError(
                        document_id=json_document["_id"], error=str(e)
                    )
                )
        return inspection_result

    class Config:
        json_encoders = {
            ObjectId: lambda v: str(v),
        }
        allow_population_by_field_name = True
        fields = {"id": "_id"}



from pydantic.main import BaseModel
from beanie import Document, PydanticObjectId
from typing import Any, Dict, Optional, List, Union


class Test(Document):
    t: str


async def get_test() -> Union[Test, Dict[str, Any]]:
    return (await Test.find().to_list())[0]


async def d(t: Test):
    print(t)


async def main():
    t = await get_test()
    await d(t)


class C:
    @classmethod
    def c(cls):
        return cls()


class B(C):
    ...


class A:
    def __init__(self, foo: Type):
        self.foo = foo

    def bar(self):
        return self.foo



from typing import (
    Union,
    Optional,
    List,
    Tuple,
    Type,
    Mapping,
    TYPE_CHECKING,
    TypeVar,
    Dict,
    Any,
)

from beanie.exceptions import DocumentNotFound
from beanie.odm.enums import SortDirection
from beanie.odm.interfaces.aggregate import AggregateMethods
from beanie.odm.interfaces.session import SessionMethods
from beanie.odm.interfaces.update import (
    UpdateMethods,
)
from beanie.odm.operators.find.logical import And
from beanie.odm.queries.aggregation import AggregationQuery
from beanie.odm.queries.cursor import BaseCursorQuery
from beanie.odm.queries.delete import (
    DeleteQuery,
    DeleteMany,
    DeleteOne,
)
from beanie.odm.queries.update import (
    UpdateQuery,
    UpdateMany,
    UpdateOne,
)
from beanie.odm.utils.projection import get_projection
from pydantic import BaseModel
from pymongo.client_session import ClientSession
from pymongo.results import UpdateResult

if TYPE_CHECKING:
    from beanie.odm.documents import DocType

FindQueryType = TypeVar("FindQueryType", bound="FindQuery")


class FindQuery(UpdateMethods, SessionMethods):
    """
    Find Query base class

    Inherited from:

    - [SessionMethods](https://roman-right.github.io/beanie/api/interfaces/#sessionmethods)
    - [UpdateMethods](https://roman-right.github.io/beanie/api/interfaces/#aggregatemethods)
    """

    UpdateQueryType: Union[
        Type[UpdateQuery], Type[UpdateMany], Type[UpdateOne]
    ] = UpdateQuery
    DeleteQueryType: Union[
        Type[DeleteOne], Type[DeleteMany], Type[DeleteQuery]
    ] = DeleteQuery

    def __init__(self, document_model: Type["DocType"]):
        self.document_model: Type["DocType"] = document_model
        self.find_expressions: List[
            Union[Dict[str, Any], Mapping[str, Any]]
        ] = []
        self.projection_model: Type[BaseModel] = self.document_model
        self.session = None

    def get_filter_query(self) -> Mapping[str, Any]:
        if self.find_expressions:
            return And(*self.find_expressions)
        else:
            return {}

    def update(
        self,
        *args: Union[Dict[str, Any], Mapping[str, Any]],
        session: Optional[ClientSession] = None
    ):
        """
        Create Update with modifications query
        and provide search criteria there

        :param args: *Union[dict, Mapping] - the modifications to apply.
        :param session: Optional[ClientSession]
        :return: UpdateMany query
        """
        self.set_session(session=session)
        return (
            self.UpdateQueryType(
                document_model=self.document_model,
                find_query=self.get_filter_query(),
            )
            .update(*args)
            .set_session(session=self.session)
        )

    def delete(
        self, session: Optional[ClientSession] = None
    ) -> Union[DeleteOne, DeleteMany]:
        """
        Provide search criteria to the Delete query

        :param session: Optional[ClientSession]
        :return: Union[DeleteOne, DeleteMany]
        """
        self.set_session(session=session)
        return self.DeleteQueryType(
            document_model=self.document_model,
            find_query=self.get_filter_query(),
        ).set_session(session=session)

    def project(
        self: FindQueryType,
        projection_model: Optional[Type[BaseModel]],
    ) -> FindQueryType:
        """
        Apply projection parameter

        :param projection_model: Optional[Type[BaseModel]] - projection model
        :return: self
        """
        if projection_model is not None:
            self.projection_model = projection_model
        return self

    def get_projection_model(self) -> Type[BaseModel]:
        return self.projection_model


class FindMany(FindQuery, BaseCursorQuery, AggregateMethods):
    """
    Find Many query class

    Inherited from:

    - [FindQuery](https://roman-right.github.io/beanie/api/queries/#findquery)
    - [BaseCursorQuery](https://roman-right.github.io/beanie/api/queries/#basecursorquery) - async generator
    - [AggregateMethods](https://roman-right.github.io/beanie/api/interfaces/#aggregatemethods)

    """

    UpdateQueryType = UpdateMany
    DeleteQueryType = DeleteMany

    def __init__(self, document_model: Type["DocType"]):
        super(FindMany, self).__init__(document_model=document_model)
        self.sort_expressions: List[Tuple[str, SortDirection]] = []
        self.skip_number: int = 0
        self.limit_number: int = 0

    def find_many(
        self,
        *args: Union[Dict[str, Any], Mapping[str, Any], bool],
        skip: Optional[int] = None,
        limit: Optional[int] = None,
        sort: Union[None, str, List[Tuple[str, SortDirection]]] = None,
        projection_model: Optional[Type[BaseModel]] = None,
        session: Optional[ClientSession] = None
    ) -> "FindMany":
        """
        Find many documents by criteria

        :param args: *Union[Dict[str, Any],
        Mapping[str, Any], bool] - search criteria
        :param skip: Optional[int] - The number of documents to omit.
        :param limit: Optional[int] - The maximum number of results to return.
        :param sort: Union[None, str, List[Tuple[str, SortDirection]]] - A key
        or a list of (key, direction) pairs specifying the sort order
        for this query.
        :param projection_model: Optional[Type[BaseModel]] - projection model
        :param session: Optional[ClientSession] - pymongo session
        :return: FindMany - query instance
        """
        self.find_expressions += args
        self.skip(skip)
        self.limit(limit)
        self.sort(sort)
        self.project(projection_model)
        self.set_session(session=session)
        return self

    def find(
        self,
        *args: Union[Dict[str, Any], Mapping[str, Any], bool],
        skip: Optional[int] = None,
        limit: Optional[int] = None,
        sort: Union[None, str, List[Tuple[str, SortDirection]]] = None,
        projection_model: Optional[Type[BaseModel]] = None,
        session: Optional[ClientSession] = None
    ) -> "FindMany":
        """
        The same as `find_many(...)`
        """
        return self.find_many(
            *args,
            skip=skip,
            limit=limit,
            sort=sort,
            projection_model=projection_model,
            session=session
        )

    def sort(
        self,
        *args: Optional[
            Union[
                str, Tuple[str, SortDirection], List[Tuple[str, SortDirection]]
            ]
        ]
    ) -> "FindMany":
        """
        Add sort parameters
        :param args: Union[str, Tuple[str, SortDirection],
        List[Tuple[str, SortDirection]]] - A key or a tuple (key, direction)
        or a list of (key, direction) pairs specifying
        the sort order for this query.
        :return: self
        """
        for arg in args:
            if arg is None:
                pass
            elif isinstance(arg, list):
                self.sort(*arg)
            elif isinstance(arg, tuple):
                self.sort_expressions.append(arg)
            elif isinstance(arg, str):
                if arg.startswith("+"):
                    self.sort_expressions.append(
                        (arg[1:], SortDirection.ASCENDING)
                    )
                elif arg.startswith("-"):
                    self.sort_expressions.append(
                        (arg[1:], SortDirection.DESCENDING)
                    )
                else:
                    self.sort_expressions.append(
                        (arg, SortDirection.ASCENDING)
                    )
            else:
                raise TypeError("Wrong argument type")
        return self

    def skip(self, n: Optional[int]) -> "FindMany":
        """
        Set skip parameter
        :param n: int
        :return: self
        """
        if n is not None:
            self.skip_number = n
        return self

    def limit(self, n: Optional[int]) -> "FindMany":
        """
        Set limit parameter
        :param n: int
        :return:
        """
        if n is not None:
            self.limit_number = n
        return self

    def update_many(
        self,
        *args: Union[Dict[str, Any], Mapping[str, Any]],
        session: Optional[ClientSession] = None
    ) -> UpdateMany:
        """
        Provide search criteria to the
        [UpdateMany](https://roman-right.github.io/beanie/api/queries/#updatemany) query

        :param args: *Union[dict, Mapping] - the modifications to apply.
        :param session: Optional[ClientSession]
        :return: [UpdateMany](https://roman-right.github.io/beanie/api/queries/#updatemany) query
        """
        return self.update(*args, session=session)

    def delete_many(
        self, session: Optional[ClientSession] = None
    ) -> DeleteMany:
        """
        Provide search criteria to the [DeleteMany](https://roman-right.github.io/beanie/api/queries/#deletemany) query

        :param session:
        :return: [DeleteMany](https://roman-right.github.io/beanie/api/queries/#deletemany) query
        """
        return self.delete(session=session)

    async def count(self) -> int:
        """
        Number of found documents
        :return: int
        """
        return (
            await self.document_model.get_motor_collection().count_documents(
                self.get_filter_query()
            )
        )

    def aggregate(
        self,
        aggregation_pipeline: List[Any],
        projection_model: Optional[Type[BaseModel]] = None,
        session: Optional[ClientSession] = None,
    ) -> AggregationQuery:
        """
        Provide search criteria to the [AggregationQuery](https://roman-right.github.io/beanie/api/queries/#aggregationquery)

        :param aggregation_pipeline: list - aggregation pipeline. MongoDB doc:
        <https://docs.mongodb.com/manual/core/aggregation-pipeline/>
        :param projection_model: Type[BaseModel] - Projection Model
        :param session: Optional[ClientSession] - PyMongo session
        :return:[AggregationQuery](https://roman-right.github.io/beanie/api/queries/#aggregationquery)
        """
        self.set_session(session=session)
        return AggregationQuery(
            aggregation_pipeline=aggregation_pipeline,
            document_model=self.document_model,
            projection_model=projection_model,
            find_query=self.get_filter_query(),
        ).set_session(session=self.session)

    @property
    def motor_cursor(self):
        return self.document_model.get_motor_collection().find(
            filter=self.get_filter_query(),
            sort=self.sort_expressions,
            projection=get_projection(self.projection_model),
            skip=self.skip_number,
            limit=self.limit_number,
            session=self.session,
        )


class FindOne(FindQuery):
    """
    Find One query class

    Inherited from:

    - [FindQuery](https://roman-right.github.io/beanie/api/queries/#findquery)
    """

    UpdateQueryType = UpdateOne
    DeleteQueryType = DeleteOne

    def find_one(
        self,
        *args: Union[Dict[str, Any], Mapping[str, Any], bool],
        projection_model: Optional[Type[BaseModel]] = None,
        session: Optional[ClientSession] = None
    ) -> "FindOne":
        """
        Find one document by criteria

        :param args: *Union[Dict[str, Any], Mapping[str, Any],
        bool] - search criteria
        :param projection_model: Optional[Type[BaseModel]] - projection model
        :param session: Optional[ClientSession] - pymongo session
        :return: FindOne - query instance
        """
        self.find_expressions += args
        self.project(projection_model)
        self.set_session(session=session)
        return self

    def update_one(
        self,
        *args: Union[Dict[str, Any], Mapping[str, Any]],
        session: Optional[ClientSession] = None
    ) -> UpdateOne:
        """
        Create [UpdateOne](https://roman-right.github.io/beanie/api/queries/#updateone) query using modifications and
        provide search criteria there
        :param args: *Union[dict, Mapping] - the modifications to apply
        :param session: Optional[ClientSession] - PyMongo sessions
        :return: [UpdateOne](https://roman-right.github.io/beanie/api/queries/#updateone) query
        """
        return self.update(*args, session=session)

    def delete_one(self, session: Optional[ClientSession] = None) -> DeleteOne:
        """
        Provide search criteria to the [DeleteOne](https://roman-right.github.io/beanie/api/queries/#deleteone) query
        :param session: Optional[ClientSession] - PyMongo sessions
        :return: [DeleteOne](https://roman-right.github.io/beanie/api/queries/#deleteone) query
        """
        return self.delete(session=session)

    async def replace_one(
        self,
        document: "DocType",
        session: Optional[ClientSession] = None,
    ) -> UpdateResult:
        """
        Replace found document by provided
        :param document: Document - document, which will replace the found one
        :param session: Optional[ClientSession] - PyMongo session
        :return: UpdateResult
        """
        self.set_session(session=session)
        result: UpdateResult = (
            await self.document_model.get_motor_collection().replace_one(
                self.get_filter_query(),
                document.dict(by_alias=True, exclude={"id"}),
                session=self.session,
            )
        )

        if not result.raw_result["updatedExisting"]:
            raise DocumentNotFound
        return result

    def __await__(self):
        """
        Run the query
        :return: BaseModel
        """
        projection = get_projection(self.projection_model)
        document: Dict[str, Any] = (
            yield from self.document_model.get_motor_collection().find_one(
                filter=self.get_filter_query(),
                projection=projection,
                session=self.session,
            )
        )
        if document is None:
            return None
        return self.projection_model.parse_obj(document)



import motor


class DBHandler:
    data = {}

    def set_db(self, uri, db_name):
        self.data["client"] = motor.motor_asyncio.AsyncIOMotorClient(uri)
        self.data["database"] = self.data["client"][db_name]

    def get_cli(self):
        return self.data.get("client")

    def get_db(self):
        return self.data.get("database")



from bson import ObjectId
from bson.errors import InvalidId
from pydantic.json import ENCODERS_BY_TYPE
from pymongo import ASCENDING

from beanie.odm.enums import SortDirection
from beanie.odm.operators.find.comparison import (
    Eq,
    GT,
    GTE,
    LT,
    LTE,
    NE,
)


def Indexed(typ, index_type=ASCENDING, **kwargs):
    """
    Returns a subclass of `typ` with an extra attribute `_indexed` as a tuple:
    - Index 0: `index_type` such as `pymongo.ASCENDING`
    - Index 1: `kwargs` passed to `IndexModel`
    When instantiated the type of the result will actually be `typ`.
    """

    class NewType(typ):
        _indexed = (index_type, kwargs)

        def __new__(cls, *args, **kwargs):
            return typ.__new__(typ, *args, **kwargs)

    NewType.__name__ = f"Indexed {typ.__name__}"
    return NewType


class PydanticObjectId(ObjectId):
    """
    Object Id field. Compatible with Pydantic.
    """

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if isinstance(v, bytes):
            v = v.decode("utf-8")
        try:
            return PydanticObjectId(v)
        except InvalidId:
            raise TypeError("Id must be of type PydanticObjectId")

    @classmethod
    def __modify_schema__(cls, field_schema):
        field_schema.update(
            type="string",
            examples=["5eb7cf5a86d9755df3a6c593", "5eb7cfb05e32e07750a1756a"],
        )


ENCODERS_BY_TYPE[
    PydanticObjectId
] = str  # it is a workaround to force pydantic make json schema for this field


class ExpressionField(str):
    def __getattr__(self, item):
        """
        Get sub field

        :param item: name of the subfield
        :return: ExpressionField
        """
        return ExpressionField(f"{self}.{item}")

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return Eq(field=self, other=other)

    def __gt__(self, other):
        return GT(field=self, other=other)

    def __ge__(self, other):
        return GTE(field=self, other=other)

    def __lt__(self, other):
        return LT(field=self, other=other)

    def __le__(self, other):
        return LTE(field=self, other=other)

    def __ne__(self, other):
        return NE(field=self, other=other)

    def __pos__(self):
        return self, SortDirection.ASCENDING

    def __neg__(self):
        return self, SortDirection.DESCENDING



from abc import abstractmethod
from typing import Optional, List, Union, Type, Dict, Any

from pydantic.main import BaseModel


class BaseCursorQuery:
    """
    BaseCursorQuery class. Wrapper over AsyncIOMotorCursor,
    which parse result with model
    """

    @abstractmethod
    def get_projection_model(self) -> Optional[Type[BaseModel]]:
        ...

    @property
    @abstractmethod
    def motor_cursor(self):
        ...

    def __aiter__(self):
        return self

    async def __anext__(self) -> Union[BaseModel, Dict[str, Any]]:
        if getattr(self, "cursor", None) is None:
            self.cursor = self.motor_cursor
        next_item = await self.cursor.__anext__()
        projection = self.get_projection_model()
        return (
            projection.parse_obj(next_item)
            if projection is not None
            else next_item
        )

    async def to_list(
        self, length: Optional[int] = None
    ) -> Union[List[BaseModel], List[Dict[str, Any]]]:  # noqa
        """
        Get list of documents

        :param length: Optional[int] - length of the list
        :return: Union[List[BaseModel], List[Dict[str, Any]]]
        """
        motor_list: List[Dict[str, Any]] = await self.motor_cursor.to_list(
            length
        )
        projection = self.get_projection_model()
        if projection is not None:
            return [projection.parse_obj(i) for i in motor_list]
        return motor_list



from typing import List, Optional, Type

from motor.motor_asyncio import AsyncIOMotorCollection, AsyncIOMotorDatabase
from pydantic.main import BaseModel
from pymongo import IndexModel


class IndexModelField(IndexModel):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if isinstance(v, IndexModel):
            return v
        else:
            return IndexModel(v)


class CollectionInputParameters(BaseModel):
    name: str = ""
    indexes: List[IndexModelField] = []

    class Config:
        arbitrary_types_allowed = True


async def collection_factory(
    database: AsyncIOMotorDatabase,
    document_model: Type,
    allow_index_dropping: bool,
    collection_class: Optional[Type] = None,
) -> Type:
    """
    Collection factory.
    Creates internal CollectionMeta class for the Document on the init step,

    :param database: AsyncIOMotorDatabase - Motor database instance
    :param document_model: Type - a class, inherited from Document class
    :param allow_index_dropping: bool - if index dropping is allowed
    :param collection_class: Optional[Type] - Collection, which was set up
    by user
    :return: Type - Collection class
    """
    # parse collection parameters
    if collection_class:
        collection_parameters = CollectionInputParameters.parse_obj(
            vars(collection_class)
        )
    else:
        collection_parameters = CollectionInputParameters()

    # set collection name
    if not collection_parameters.name:
        collection_parameters.name = document_model.__name__

    # create motor collection
    collection = database[collection_parameters.name]

    # indexes
    old_indexes = (await collection.index_information()).keys()
    new_indexes = ["_id_"]

    # Indexed field wrapped with Indexed()
    found_indexes = [
        IndexModel(
            [(fname, fvalue.type_._indexed[0])], **fvalue.type_._indexed[1]
        )
        for fname, fvalue in document_model.__fields__.items()
        if hasattr(fvalue.type_, "_indexed") and fvalue.type_._indexed
    ]

    # get indexes from the Collection class
    if collection_parameters.indexes:
        found_indexes += collection_parameters.indexes

    # create indices
    if found_indexes:
        new_indexes += await collection.create_indexes(found_indexes)

    # delete indexes
    if allow_index_dropping:
        for index in set(old_indexes) - set(new_indexes):
            await collection.drop_index(index)

    # create internal CollectionMeta class for the Document
    class CollectionMeta:
        name: str = collection_parameters.name
        motor_collection: AsyncIOMotorCollection = collection
        indexes: List = found_indexes

    return CollectionMeta



from abc import abstractmethod
from typing import Any, Dict, Union, Mapping

from beanie.odm.operators import BaseOperator


class BaseUpdateOperator(BaseOperator):
    @property
    @abstractmethod
    def query(self) -> Union[Dict[str, Any], Mapping[str, Any]]:
        ...



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['beanie',
 'beanie.executors',
 'beanie.migrations',
 'beanie.migrations.controllers',
 'beanie.odm',
 'beanie.odm.interfaces',
 'beanie.odm.operators',
 'beanie.odm.operators.find',
 'beanie.odm.operators.update',
 'beanie.odm.queries',
 'beanie.odm.utils']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.2,<8.0.0',
 'motor>=2.1.0,<3.0.0',
 'pydantic>=1.8,<2.0',
 'toml>=0.10.2,<0.11.0']

entry_points = \
{'console_scripts': ['beanie = beanie.executors.migrate:migrations']}

setup_kwargs = {
    'name': 'beanie',
    'version': '1.1.2',
    'description': 'Asynchronous Python ODM for MongoDB',
    'long_description': '[![Beanie](https://raw.githubusercontent.com/roman-right/beanie/main/assets/logo/with_text.svg)](https://github.com/roman-right/beanie)\n\n## Overview\n\n[Beanie](https://github.com/roman-right/beanie) - is an Asynchronous Python\nobject-document mapper (ODM) for MongoDB, based\non [Motor](https://motor.readthedocs.io/en/stable/)\nand [Pydantic](https://pydantic-docs.helpmanual.io/).\n\nWhen using Beanie each database collection has a corresponding `Document` that\nis used to interact with that collection. In addition to retrieving data,\nBeanie allows you to add, update, or delete documents from the collection as\nwell.\n\nBeanie saves you time by removing boiler-plate code and it helps you focus on\nthe parts of your app that actually matter.\n\nData and schema migrations are supported by Beanie out of the box.\n\n## Installation\n\n### PIP\n\n```shell\npip install beanie\n```\n\n### Poetry\n\n```shell\npoetry add beanie\n```\n\n## Getting Started\n\n### Document\n\nThe `Document` class in Beanie is responsible for mapping and handling the data\nfrom the collection. It is inherited from the `BaseModel` Pydantic class, so it\nfollows the same data typing and parsing behavior.\n\nEach document has `id` fields of type `PydanticObjectId` (Can be changed, if you prefer another type) which reflects the\nunique `_id` field in\nMongoDB. [MongoDB doc](https://docs.mongodb.com/manual/reference/glossary/#term-id)\n\n```python\nfrom typing import Optional\nfrom pydantic import BaseModel\nfrom beanie import Document, Indexed\n\n\nclass Category(BaseModel):\n    name: str\n    description: str\n\n\nclass Product(Document):  # This is the model\n    name: str\n    description: Optional[str] = None\n    price: Indexed(float)\n    category: Category\n\n    class Collection:\n        name = "products"\n```\n\nMore details about Documents, collections, and indexes configuration could be\nfound in the [documentation](https://roman-right.github.io/beanie/).\n\n### Initialization\n\nBeanie uses Motor as an async database engine. To init previously created\ndocuments, you should provide the Motor database instance and list of your\ndocument models to the `init_beanie(...)` function, as it is shown in the\nexample:\n\n```python\nimport motor\nfrom beanie import init_beanie\n\n\nasync def init():\n    # Crete Motor client\n    client = motor.motor_asyncio.AsyncIOMotorClient(\n        "mongodb://user:pass@host:27017"\n    )\n\n    # Init beanie with the Product document class\n    await init_beanie(database=client.db_name, document_models=[Product])\n\n```\n\n### Create\n\nBeanie supports and single document creation and batch inserts:\n\n#### Insert one\n\n```python\nbar = Product(name="Tony\'s", price=5.95, category=Category(name="Chocolate"))\nawait bar.insert()\n```\n\n#### Insert many\n\n```python\nmilka = Product(name="Milka", price=3.05, category=chocolate)\npeanut_bar = Product(name="Peanut Bar", price=4.44, category=chocolate)\nawait Product.insert_many([milka, peanut_bar])\n```\n\nOther details and examples could be found in the [documentation](https://roman-right.github.io/beanie/insert/)\n\n### Find\n\n#### Get\n\nThe simplest way to get the document is to get it by the `id` field.\nMethod `get()` is responsible for this:\n\n```python\nbar = await Product.get(PydanticObjectId("608da169eb9e17281f0ab2ff"))\n```\n\n#### Find One\n\nTo get a single document by any other search criteria you can use `find_one()`\nmethod. It responds with awaitable object `FindOne`, which returns `Document`\ninstance or `None` on `await`\n\n```python\nbar = await Product.find_one(\n    Product.name == "Peanut Bar",\n    Product.category.name == "Chocolate"\n)\n```\n\n`FindOne` supports projections to grab and return data in needed format only\n\n```python\nclass ProductShortView(BaseModel):\n    name: str\n    price: float\n\n\nbar = await Product.find_one(\n    Product.name == "Peanut Bar"\n).project(ProductShortView)\n```\n\n#### Find Many\n\nTo find many documents `find` or `find_many` (which is the same) method could\nbe used. These methods return the `FindMany` instance, which implements\nthe `async generator` pattern. It means found documents are available in\nthe `async for` loop:\n\n```python\nasync for product in Product.find(\n        Product.category.name == "Chocolate"\n):\n    print(product)\n```\n\nThe `to_list` method will put all the found documents on the list\n\n```python\nproducts = await Product.find(\n    Product.category.name == "Chocolate"\n).to_list()\n```\n\n`FindMany` supports chain filtering, where you can build your filter query with\nmany `find` methods\n\n```python\nchocolates = await Product.find(\n    Product.category.name == "Chocolate").find(\n    Product.price < 5).to_list()\n```\n\nYou can sort, skip, limit and project with `FindMany` query too\n\n```python\nclass ProductShortView(BaseModel):\n    name: str\n    price: float\n\n\nproducts = await Product.find(\n    Product.category.name == "Chocolate",\n    Product.price < 3.5\n).sort(-Product.price).limit(10).project(ProductShortView)\n\n```\n\n#### Query Building\n\nAnd `FindOne` and `FindMany` support native python comparison operators, Beanie\nfind operators and native PyMongo find query syntax.\n\n**Python comparison operators**\n\n```python\nchocolates = await Product.find(\n    Product.category.name == "Chocolate").find(\n    Product.price < 5).to_list()\n```\n\n**Beanie Find Operators** are classes - wrappers over MongoDB find operators.\n\n```python\nfrom beanie.operators import In\n\nproducts = await Product.find(\n    In(Product.price, [1, 2, 3, 4, 5])\n).to_list()\n```\n\nHere you can see `In` operator, which reflects `$in`. The whole list of the\nfind operators could be\nfound [here](https://roman-right.github.io/beanie/api/operators/find/)\n\n**Native PyMongo find query syntax**\n\n```python\nproducts = await Product.find(\n    {"$in": {"price": [1, 2, 3, 4, 5]}}\n).to_list()\n```\n\n#### All\n\nTo get all the documents `find_all` or `all` methods can be used.\n\n```python\nall_products = await Product.all().to_list()\n```\n\nInformation about sorting, skips, limits, and projections could be found in\nthe [documentation](https://roman-right.github.io/beanie/find/)\n\n### Update\n\n#### Saving changes to existing documents\n\nThe easiest way to change a document in the database is using either the `replace` or `save` methods on a altered document. These methods both write the document to the database, but `replace` will raise an exception when the document does not exist yet, while `save` will insert the document. \n\nUsing save:\n```python\nbar = await Product.find_one(Product.name == "Milka")\nbar.price = 10\nawait bar.save()\n```\nOr similairly using replace, which trows a `ValueError` if the document does not have an `id` yet, or a `beanie.exceptions.DocumentNotFound` if it does but the id is not present in the collection:\n```python\nbar.price = 10\ntry:\n    await bar.update()\nexcept (ValueError, beanie.exceptions.DocumentNotFound):\n    print("Can\'t replace a non existing document")\n```\n\nNote however that these methods requires multiple queries to the database and replace the entire document with the new version. A more tailered solution can often be created by applying update queries directly on the database level, see the explination below.\n\n#### Update Methods\n\n`Document`, `FindMany` and `FindOne`\nimplement [UpdateMethods](https://roman-right.github.io/beanie/api/interfaces/#updatemethods)\ninterface. It supports `update`, `set`, `inc` and `current_date` methods.\n\n#### Document\n\nImplementing `UpdateMethods` interface `Document` instance\ncreates [UpdateOne](https://roman-right.github.io/beanie/api/queries/#updateone)\nobject and provides self id as the search criteria there. Then `UpdateOne`\ninstance is handling all the update operations.\n\n`update` method is used to update the document data. It supports native Pymongo\nsyntax and Beanie Update operators.\n\nNative syntax\n\n```python\nbar = await Product.find_one(Product.name == "Milka")\nawait bar.update({"$set": {Product.price: 10}})\n```\n\nUpdate operatos\n\n```python\nawait bar.update(Set({Product.price: 10}))\n```\n\nThe whole list of the Beanie update operators can be\nfound [by link](https://roman-right.github.io/beanie/api/operators/update/)\n\n`inc`, `set` and `current_date` methods are popular update operations preset.\nNext example shows how to add `1` to the document\'s price:\n\n```python\nawait bar.inc({Product.price: 1})\n```\n\n#### Update One\n\nImplementing `UpdateMethods` interface `Document` instance\ncreates [UpdateOne](https://roman-right.github.io/beanie/api/queries/#updateone)\nobject and provides search criteria there. All the update methods work the same\nway as for the `Document` instance.\n\nNative syntax\n\n```python\nawait Product.find_one(\n    Product.name == "Milka"\n).update({"$set": {Product.price: 10}})\n```\n\nUpdate Operators\n\n```python\nawait Product.find_one(\n    Product.name == "Milka"\n).update(Set({Product.price: 10}))\n```\n\nPreset Methods\n\n```python\nawait Product.find_one(\n    Product.name == "Milka"\n).inc({Product.price: 1})\n```\n\n#### Update Many\n\n`FindMany` uses the same patter as `FindOne` but creates `UpdateMany` instance\ninstead of `UpdateOne` respectively. It supports `UpdateMethods` too.\n\n```python\nawait Product.find(\n    Product.category.name == "Chocolate"\n).update({"$set": {Product.price: 100}})\n```\n\nMore details and examples about update queries could be found in\nthe [documentation](https://roman-right.github.io/beanie/update/)\n\n### Delete\n\n`delete()` method is supported and by the `Document` instances, and by the `FindOne`\nand by the `FindMany` instances. It deletes documents using id or search\ncriteria respectively.\n\n#### Document\n\n```python\nbar = await Product.find_one(Product.name == "Milka")\nawait bar.delete()\n```\n\n#### One\n\n```python\nawait Product.find_one(Product.name == "Milka").delete()\n```\n\n#### Many\n\n```python\nawait Product.find(\n    Product.category.name == "Chocolate"\n).delete()\n```\n\n#### ALL\n\n```python\nawait Product.delete_all()\n```\n\nMore information could be found in the [documentation](https://roman-right.github.io/beanie/delete/)\n\n### Aggregate\n\nYou can aggregate and over the whole collection, using `aggregate()` method of the `Document` class, and over search criteria, using `FindMany` instance. \n\n#### Aggregation Methods\n\n`FindMany` and `Document` classes implements [AggregateMethods](https://roman-right.github.io/beanie/api/interfaces/#aggregatemethods) interface with preset methods\n\nExample of average calculation:\n\n*With search criteria*\n```python\navg_price = await Product.find(\n    Product.category.name == "Chocolate"\n).avg(Product.price)\n```\n\n*Over the whole collection*\n```python\navg_price = await Product.avg(Product.price)\n```\n\n#### Native syntax\n\nYou can use the native PyMongo syntax of the aggregation pipelines to aggregate over the whole collection or over the subset too. `projection_model` parameter is responsible for the output format. It will return dictionaries, if this parameter is not provided.\n\n```python\nclass OutputItem(BaseModel):\n    id: str = Field(None, alias="_id")\n    total: int\n\n\nresult = await Product.find(\n    Product.category.name == "Chocolate").aggregate(\n    [{"$group": {"_id": "$category.name", "total": {"$avg": "$price"}}}],\n    projection_model=OutputItem\n).to_list()\n\n```\n\n### Documentation\n\n- **[Doc](https://roman-right.github.io/beanie/)** -\n  Usage examples with descriptions\n- **[API](https://roman-right.github.io/beanie/api/document/)** - Full list of\n  the classes and methods\n\n### Example Projects\n\n- **[FastAPI Demo](https://github.com/roman-right/beanie-fastapi-demo)** -\n  Beanie and FastAPI collaboration demonstration. CRUD and Aggregation.\n- **[Indexes Demo](https://github.com/roman-right/beanie-index-demo)** -\n  Regular and Geo Indexes usage example wrapped to a microservice.\n\n### Articles\n\n- **[Announcing Beanie - MongoDB ODM](https://dev.to/romanright/announcing-beanie-mongodb-odm-56e)**\n- **[Build a Cocktail API with Beanie and MongoDB](https://developer.mongodb.com/article/beanie-odm-fastapi-cocktails/)**\n- **[MongoDB indexes with Beanie](https://dev.to/romanright/mongodb-indexes-with-beanie-43e8)**\n- **[Beanie Projections. Reducing network and database load.](https://dev.to/romanright/beanie-projections-reducing-network-and-database-load-3bih)**\n\n### Resources\n\n- **[GitHub](https://github.com/roman-right/beanie)** - GitHub page of the\n  project\n- **[Changelog](https://roman-right.github.io/beanie/changelog)** - list of all\n  the valuable changes\n- **[Discord](https://discord.gg/ZTTnM7rMaz)** - ask your questions, share\n  ideas or just say `Hello!!`\n\n----\nSupported by [JetBrains](https://jb.gg/OpenSource)\n\n[![JetBrains](https://raw.githubusercontent.com/roman-right/beanie/main/assets/logo/jetbrains.svg)](https://jb.gg/OpenSource)\n',
    'author': 'Roman',
    'author_email': 'roman-right@protonmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/roman-right/beanie',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4.0',
}


setup(**setup_kwargs)
